#!/usr/bin/python

from datetime import datetime

import time
import random
import socket
import sys
import threading
import codecs

def gendevice(devtype, host, mac):
  if devtype == 0x0202: # HVAC
    return hvac(host=host, mac=mac)
  else:
    return device(host=host, mac=mac)

def discover(timeout=None, local_ip_address=None):
  if local_ip_address is None:
      s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
      s.connect(('8.8.8.8', 53))  # connecting to a UDP address doesn't send packets
      local_ip_address = s.getsockname()[0]
  address = local_ip_address.split('.')
  cs = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  cs.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
  cs.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
  cs.bind((local_ip_address,0))
  port = cs.getsockname()[1]
  starttime = time.time()

  devices = []

  timezone = int(time.timezone/-3600)
  packet = bytearray.fromhex('AAAA0602FFFFFF0059')

  cs.sendto(packet, ('255.255.255.255', 12414))

  if timeout is None:
    response = cs.recvfrom(1024)
    responsepacket = bytearray(response[0])
    host = response[1]
    mac = responsepacket[0x7:0x0e]
    devtype = responsepacket[0x5] | responsepacket[0x6] << 8
    cs.close()
    return gendevice(devtype, host, mac)
  else:
    while (time.time() - starttime) < timeout:
      cs.settimeout(timeout - (time.time() - starttime))
      try:
        response = cs.recvfrom(1024)
      except socket.timeout:
        return devices
      responsepacket = bytearray(response[0])
      host = response[1]
      devtype = responsepacket[0x05] | responsepacket[0x06] << 8
      mac = responsepacket[0x07:0x0E]
      dev = gendevice(devtype, host, mac)
      devices.append(dev)
    cs.close()
    return devices

class device:
  def __init__(self, host, mac, timeout=10):
    self.host = host
    self.mac = mac
    self.timeout = timeout
    self.count = random.randrange(0xffff)
    self.type = "Unknown"
    self.lock = threading.Lock()

  def auth(self):
    return True

  def checksum(self, payload):
    checksum = bytearray.fromhex(hex(sum(payload))[3:].upper())
    payload.extend(checksum)
    return payload

  def send_packet(self, payload):
    packet=payload
    starttime = time.time()
    ts = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ts.connect((self.host[0], 12416))

    with self.lock:
      while True:
        try:
          ts.send(packet)
          ts.settimeout(1)
          response = ts.recv(2048)
          break
        except socket.timeout:
          if (time.time() - starttime) > self.timeout:
            raise
    ts.close()

    return bytearray(response)

class hvac(device):
  def __init__ (self, host, mac):
    device.__init__(self, host, mac)
    self.type = "HVAC Smart WiFi"

  def send_request(self,input_payload):
    request_payload = self.checksum(input_payload)
    response_payload = self.send_packet(request_payload)
    response_payload_len = response_payload[0]

    # check for error
    check_payload = self.checksum(response_payload[0:(len(response_payload)-1)])
    if response_payload != check_payload:
#	print ("CRC ERROR")
    return response_payload[0:len(response_payload)]

  def calc_time(self, tm1 , tm2):
    tmr = int((bin(tm1)[2:] + bin(tm2)[2:].zfill(8)),2)
    hrs = int(tmr / 60)
    mins = int(tmr % 60)
    return (hrs, mins)

  def get_current_status(self):
    payload = self.send_request(bytearray([0xAA,0xAA,0x12,0xA0,0x0A,0x0A,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00]))
    return payload[0:len(payload)-1]

  # Get full status
  def get_full_status(self):
    payload = self.get_current_status()
    data = {}
    data['ac_mode'] =  payload[7] & 7
    data['power'] =  (payload[7] >> 3) & 1
    data['fan_speed'] =  str(int(payload[7]) >> 4)
    data['temperature'] =  ((payload[8] & 15) + 16)
    data['temptype'] =  (payload[8] >> 5) & 1
    data['quiet'] =  (payload[8] >> 6) & 1
    data['fan_direction'] =  payload[9] & 15
    data['eco'] =  payload[10] & 1
    data['light'] =  (payload[10] >> 7) & 1
    data['health'] =  (payload[10] >> 6) & 1
    data['timing'] =  (payload[10] >> 5) & 1
    data['dry'] =  (payload[10] >> 4) & 1
    data['wdnumber_mode'] =  (payload[10] >> 2) & 3
    data['sleep'] =  (payload[10] >> 1) & 1
    data['energy save'] =  (payload[10] >> 4) & 1
    tm1=payload[11]
    if (tm1 & 0xF):
        tm2 = payload[12]
        tm1 = tm1 & 7
        tms = self.calc_time (tm1, tm2)
        data['timer'] = 1
        data['timer_h'] = tms[0]
        data['timer_m'] = tms[1]
    elif (tm1 & 0xF0):
        tm2 = payload[13]
        tm1 = (tm1 >> 4) & 7
        tms = self.calc_time (tm1, tm2)
        data['timer'] = 2
        data['timer_h'] = tms[0]
        data['timer_m'] = tms[1]
    else:
        data['timer'] = 0
        data['timer_h'] = 0
        data['timer_m'] = 0
    data['stepless_max'] =  payload[14]
    data['indoorTemperature'] =  (str(payload[15])+"."+str(payload[16]))
    return data

  def set_temp(self, temperature):
    input_payload = self.get_current_status()
    input_payload[8] = (input_payload[8] & 240) | (temperature - 16)
    input_payload[3] = (input_payload[3] | 1) 
    payload = self.send_request(input_payload)

  def set_power(self, mode):
    input_payload = self.get_current_status()
    input_payload[7] = input_payload[7] & 247
    if (mode):
      input_payload[7] = input_payload[7] | 8
    input_payload[3] = (input_payload[3] | 1) 
    payload = self.send_request(input_payload)

  def set_ac_mode(self, mode):
    input_payload = self.get_current_status()
    input_payload[7] = input_payload[7] & 248
    input_payload[7] = input_payload[7] | mode
    input_payload[3] = (input_payload[3] | 1)
    payload = self.send_request(input_payload)

  def set_fan_speed(self, mode):
    input_payload = self.get_current_status()
    input_payload[7] = input_payload[7] & 15
    result_mode = {
      0: lambda : 0,
      1: lambda : 16,
      2: lambda : 32,
      3: lambda : 48,
      4: lambda : 64,
      5: lambda : 80,
      6: lambda : 96,
      8: lambda : 128
    }[mode]()
    input_payload[7] = input_payload[7] | result_mode
    input_payload[3] = (input_payload[3] | 1)
    payload = self.send_request(input_payload)

  def set_quiet(self, mode):
    input_payload = self.get_current_status()
    input_payload[8] = input_payload[8] & 191
    if (mode):
      input_payload[8] = input_payload[8] | 64
    input_payload[3] = (input_payload[3] | 1)
    payload = self.send_request(input_payload)

  def set_fan_direction(self, mode):
    input_payload = self.get_current_status()
    input_payload[9] = input_payload[9] & 240
    input_payload[9] = input_payload[9] | mode
    input_payload[3] = (input_payload[3] | 1)
    payload = self.send_request(input_payload)

  def set_stepless_max(self, stepless_max):
    input_payload = self.get_current_status()
    input_payload[14] = stepless_max
    input_payload[3] = (input_payload[3] | 1)
    payload = self.send_request(input_payload)

  def set_light(self, mode):
    input_payload = self.get_current_status()
    input_payload[10] = input_payload[10] & 127
    if (mode):
      input_payload[10] = input_payload[10] | 128
    input_payload[3] = (input_payload[3] | 1) 
    payload = self.send_request(input_payload)

  def set_health(self, mode):
    input_payload = self.get_current_status()
    input_payload[10] = input_payload[10] & 191
    if (mode):
      input_payload[10] = input_payload[10] | 64
    input_payload[3] = (input_payload[3] | 1) 
    payload = self.send_request(input_payload)

  def set_sleep(self, mode):
    input_payload = self.get_current_status()
    input_payload[10] = input_payload[10] & 253
    if (mode):
      input_payload[10] = input_payload[10] | 2
    input_payload[3] = (input_payload[3] | 1) 
    payload = self.send_request(input_payload)

  def set_energy_save(self, mode):
    input_payload = self.get_current_status()
    input_payload[10] = input_payload[10] & 239
    if (mode):
      input_payload[10] = input_payload[10] | 16
    input_payload[3] = (input_payload[3] | 1) 
    payload = self.send_request(input_payload)

  def reset_timer(self):
    input_payload = self.get_current_status()
    tm1=input_payload[11]
    if ((tm1 & 15) != 0):
	tm1 = tm1 & 7
	input_payload[11] = tm1
        input_payload[3] = (input_payload[3] | 1) 
	payload = self.send_request(input_payload)
    elif (tm1 & 240 != 0):
	tm1 = tm1 & 112
	input_payload[11] = tm1
        input_payload[3] = (input_payload[3] | 1) 
	payload = self.send_request(input_payload)

  def set_timer(self, tmh , tmm, tmst):
    self.reset_timer()
    input_payload = self.get_current_status()
    if ((tmst == 1) and (((input_payload[7] >> 3) & 1) != 1)):
        tmr = bytearray.fromhex(hex((int(bin((tmh * 60) + tmm),2) | 2048))[2:].zfill(4))
        input_payload[11] = int(tmr[0])
        input_payload[12] = int(tmr[1])
        input_payload[3] = (input_payload[3] | 1) 
	payload = self.send_request(input_payload)
    elif ((tmst == 2) and (((input_payload[7] >> 3) & 1) ==1)):
        tmr = bytearray.fromhex(hex((int(bin((tmh * 60) + tmm),2) | 2048))[2:].zfill(4))
        tmr[0] = tmr[0] << 4
	input_payload[11] = int(tmr[0])
	input_payload[13] = int(tmr[1])
        input_payload[3] = (input_payload[3] | 1) 
	payload = self.send_request(input_payload)

